/*
 *	matrix.cpp
 *	Homogenous 3D transformation matrix
 */

#include <iostream>
#include <iomanip>
#include <string>
#include <cmath>

#include "matrix.h"
#include "vector.h"
using namespace std;
/*
 * constructor
 */
matrix::matrix( 
		float v_11, float v_12, float v_13, float v_14,
		float v_21, float v_22, float v_23, float v_24,
		float v_31, float v_32, float v_33, float v_34,
		float v_41, float v_42, float v_43, float v_44
) {
	data[_11] = v_11; data[_12] = v_12; data[_13] = v_13; data[_14] = v_14; 
	data[_21] = v_21; data[_22] = v_22; data[_23] = v_23; data[_24] = v_24;
	data[_31] = v_31; data[_32] = v_32; data[_33] = v_33; data[_34] = v_34; 
	data[_41] = v_41; data[_42] = v_42; data[_43] = v_43; data[_44] = v_44; 
}

/*
 * constructor 
 */
matrix::matrix(const float v[16]) {
	memcpy(data, v, sizeof(float)*16);	
}


/*
 * copy constructor 
 */
matrix::matrix(const matrix &source) {
	memcpy(data, source.data, sizeof(float)*16);	
}

/*
 * defualt constructor 
 */
matrix::matrix() {
	identity();	
}


/*
 * identity:
 *	|1		0		0		0|
 * |0		1		0		0| 
 * |0		0		1		0|
 *	|0		0		0		1|
 */
void matrix::identity() {
	data[_11] = 1; data[_12] = 0; data[_13] = 0; data[_14] = 0; 
	data[_21] = 0; data[_22] = 1; data[_23] = 0; data[_24] = 0;
	data[_31] = 0; data[_32] = 0; data[_33] = 1; data[_34] = 0; 
	data[_41] = 0; data[_42] = 0; data[_43] = 0; data[_44] = 1; 	
}


/*
 *
 *  translate:
 *	| _11 _12 _13 _14|	|1		0		0		t_x|
 * | _21 _22 _23 _24| *	|0		1		0		t_y| 
 * | _31 _32 _33 _34|	|0		0		1		t_z|
 *	| _41 _42 _43 _44|	|0		0		0		1  |
 *
 */
void matrix::translate(float t_x, float t_y, float t_z)
{
    data[_14] = data[_11]*t_x + data[_12]*t_y + data[_13]*t_z + data[_14];
    data[_24] = data[_21]*t_x + data[_22]*t_y + data[_23]*t_z + data[_24];
    data[_34] = data[_31]*t_x + data[_32]*t_y + data[_33]*t_z + data[_34];
    data[_44] = data[_41]*t_x + data[_42]*t_y + data[_43]*t_z + data[_44];
}


/*
 *
 * scale:
 * | __11 __12 __13 _14 |		|s_x	0		0		0|
 * | __21 __22 __23 _24 |  *	|0		s_y	0		0| 
 * | __31 __32 __33 _34 |		|0		0		s_z	0|
 *	| __41 __42 __43 _44 |		|0		0		0		1|
 *
 */
void matrix::scale(float s_x, float s_y, float s_z) {
	data[_11] *=s_x;	data[_12] *=s_y;	data[_13] *=s_z;
	data[_21] *=s_x;	data[_22] *=s_y;	data[_23] *=s_z;
	data[_31] *=s_x;	data[_32] *=s_y;	data[_33] *=s_z;
	data[_41] *=s_x;	data[_42] *=s_y;	data[_43] *=s_z;
}


/*
 *
 * rotateX:
 *	angle : angle of rotation 'a' about X-axis
 *
 *	| __11 __12 __13 __14 |		|1		0			0			0|
 * | __21 __22 __23 __24 |	 *	|0		cos(a)	-sin(a)	0| 
 * | __31 __32 __33 __34 |		|0		sin(a)	cos(a)	0|
 *	| __41 __42 __43 __44 |		|0		0			0			1|
 *
 */
void matrix::rotateX(float angle) {
	float a = data[_12];
	float b = data[_22];
	float c = data[_32];
	float d = data[_42];

	angle = degr2rad(angle);
	float cosTheta = (float) cos(angle);
	float sinTheta = (float) sin(angle);
    
   data[_12] = a  * cosTheta + data[_13] * sinTheta;
   data[_13] = a  * -sinTheta + data[_13] * cosTheta;

   data[_22] = b  * cosTheta + data[_23] * sinTheta;
   data[_23] = b  * -sinTheta + data[_23] * cosTheta;

   data[_32] = c  * cosTheta + data[_33] * sinTheta;
   data[_33] = c  * -sinTheta + data[_33] * cosTheta;

   data[_42] = d  * cosTheta + data[_43] * sinTheta;
   data[_43] = d  * -sinTheta + data[_43] * cosTheta;
}

/*
 *
 * rotateY:
 *	angle : angle of rotation 'a' about Y-axis
 *
 *	| _11 _12 _13 _14 |		|cos(a)  0	sin(a)	0|
 * | _21 _22 _23 _24 |	 *	|0			1	0			0| 
 * | _31 _32 _33 _34 |		|-sin(a) 0	cos(a)	0|
 *	| _41 _42 _43 _44 |		|0			0	0			1|
 *
 */
void matrix::rotateY(float angle) {
	float a = data[_11];
	float b = data[_21];
	float c = data[_31];
	float d = data[_41];

	angle = degr2rad(angle);
	float cosTheta = (float) cos(angle);
	float sinTheta = (float) sin(angle);
    
   data[_11] = a  * cosTheta  + data[_13] * -sinTheta;
   data[_13] = a  * sinTheta + data[_13] * cosTheta;

   data[_21] = b  * cosTheta  + data[_23] * -sinTheta;
   data[_23] = b  * sinTheta + data[_23] * cosTheta;

   data[_31] = c  * cosTheta  + data[_33] * -sinTheta;
   data[_33] = c  * sinTheta + data[_33] * cosTheta;

   data[_41] = d  * cosTheta  + data[_43] * -sinTheta;
   data[_43] = d  * sinTheta + data[_43] * cosTheta;
}

/*
 *
 * rotateZ:
 *	angle : angle of rotation 'a' about Z-axis
 *
 *	| _11 _12 _13 _14 |		|cos(a)	-sin(a)	0	0|
 * | _21 _22 _23 _24 |	 *	|sin(a)	cos(a)	0	0|
 * | _31 _32 _33 _34 |		|0			0			1	0|
 *	| _41 _42 _43 _44 |		|0			0			0	1|
 *
 */
void matrix::rotateZ(float angle) {
	float a = data[_11];
	float b = data[_21];
	float c = data[_31];
	float d = data[_41];

	angle = degr2rad(angle);
	float cosTheta = (float) cos(angle);
	float sinTheta = (float) sin(angle);

   data[_11] = a  * cosTheta + data[_12] * sinTheta;
   data[_12] = a  * -sinTheta + data[_12] * cosTheta;

   data[_21] = b  * cosTheta + data[_22] * sinTheta;
   data[_22] = b  * -sinTheta + data[_22] * cosTheta;

   data[_31] = c  * cosTheta + data[_32] * sinTheta;
   data[_32] = c  * -sinTheta + data[_32] * cosTheta;

   data[_41] = d  * cosTheta + data[_42] * sinTheta;
   data[_42] = d  * -sinTheta + data[_42] * cosTheta;
}

/*
 *
 * multiply
 *
 *                  | _11 _12 _13 _14 |   | r11 r12 r13 r14 |
 *                  | _21 _22 _23 _24 |   | r21 r22 r23 r24 |
 * l' = l * r =	  | _31 _32 _33 _34 | * | r31 r32 r33 r34 |
 *                  | _41 _42 _43 _44 |   | r41 r42 r43 r44 |
 *
 */
void matrix::mult(const matrix &rm) {
   matrix lm(*this);
	const float *r = rm.data; 
	float *l = lm.data;
	

   l[_11] = data[_11]*r[_11]+data[_12]*r[_21]+data[_13]*r[_31]+data[_14]*r[_41];
   l[_12] = data[_11]*r[_12]+data[_12]*r[_22]+data[_13]*r[_32]+data[_14]*r[_42];
   l[_13] = data[_11]*r[_13]+data[_12]*r[_23]+data[_13]*r[_33]+data[_14]*r[_43];
   l[_14] = data[_11]*r[_14]+data[_12]*r[_24]+data[_13]*r[_34]+data[_14]*r[_44];

   l[_21] = data[_21]*r[_11]+data[_22]*r[_21]+data[_23]*r[_31]+data[_24]*r[_41];
   l[_22] = data[_21]*r[_12]+data[_22]*r[_22]+data[_23]*r[_32]+data[_24]*r[_42];
   l[_23] = data[_21]*r[_13]+data[_22]*r[_23]+data[_23]*r[_33]+data[_24]*r[_43];
   l[_24] = data[_21]*r[_14]+data[_22]*r[_24]+data[_23]*r[_34]+data[_24]*r[_44];

   l[_31] = data[_31]*r[_11]+data[_32]*r[_21]+data[_33]*r[_31]+data[_34]*r[_41];
   l[_32] = data[_31]*r[_12]+data[_32]*r[_22]+data[_33]*r[_32]+data[_34]*r[_42];
   l[_33] = data[_31]*r[_13]+data[_32]*r[_23]+data[_33]*r[_33]+data[_34]*r[_43];
   l[_34] = data[_31]*r[_14]+data[_32]*r[_24]+data[_33]*r[_34]+data[_34]*r[_44];

   l[_41] = data[_41]*r[_11]+data[_42]*r[_21]+data[_43]*r[_31]+data[_44]*r[_41];
   l[_42] = data[_41]*r[_12]+data[_42]*r[_22]+data[_43]*r[_32]+data[_44]*r[_42];
   l[_43] = data[_41]*r[_13]+data[_42]*r[_23]+data[_43]*r[_33]+data[_44]*r[_43];
   l[_44] = data[_41]*r[_14]+data[_42]*r[_24]+data[_43]*r[_34]+data[_44]*r[_44];
	
	memcpy(data, l, sizeof(float)*16);
}


/*
 * print
 */
void matrix::print() {
	cout<<setw(8)<<data[_11]<<", "<<setw(8)<<data[_12]<<", "<<setw(8)<<data[_13]<<", "<<setw(8)<<data[_14]<<"\n";
	cout<<setw(8)<<data[_21]<<", "<<setw(8)<<data[_22]<<", "<<setw(8)<<data[_23]<<", "<<setw(8)<<data[_24]<<"\n";
	cout<<setw(8)<<data[_31]<<", "<<setw(8)<<data[_32]<<", "<<setw(8)<<data[_33]<<", "<<setw(8)<<data[_34]<<"\n";
	cout<<setw(8)<<data[_41]<<", "<<setw(8)<<data[_42]<<", "<<setw(8)<<data[_43]<<", "<<setw(8)<<data[_44]<<endl;
}


/*
 * assignment operator  
 */
matrix& matrix::operator=(const matrix& m) {
	memcpy(data, m.getData(), sizeof(float)*16);
	return *this;
}

